﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using HeuristicLab.Data;
using HeuristicLab.Common;
using HeuristicLab.Persistence.Default.CompositeSerializers.Storable;
using HeuristicLab.Core;

namespace HeuristicLab.Problems.MetaOptimization {
  [StorableClass]
  public class DoubleValueRange : Range<DoubleValue> {

    public DoubleValueRange(DoubleValue lowerBound, DoubleValue upperBound, DoubleValue stepSize) : base(lowerBound, upperBound, stepSize) { }
    public DoubleValueRange() { }
    [StorableConstructor]
    protected DoubleValueRange(bool deserializing) : base(deserializing) { }
    protected DoubleValueRange(DoubleValueRange original, Cloner cloner) : base(original, cloner) { }
    public override IDeepCloneable Clone(Cloner cloner) {
      return new DoubleValueRange(this, cloner);
    }

    protected override DoubleValue GetRandomSample(IRandom random) {
      double val;
      do {
        val = Math.Round((random.NextDouble() * (UpperBound.Value - LowerBound.Value) + LowerBound.Value) / StepSize.Value, 0) * StepSize.Value;
      } while (val < LowerBound.Value || val > UpperBound.Value);
      return new DoubleValue(val);
    }

    public void Fix(DoubleValue value) {
      // apply stepsize
      value.Value = ((int)Math.Round(value.Value / this.StepSize.Value, 0)) * this.StepSize.Value;

      // repair bounds
      if (value.Value > this.UpperBound.Value) value.Value = this.UpperBound.Value;
      if (value.Value < this.LowerBound.Value) value.Value = this.LowerBound.Value;
    }

    public override IEnumerable<DoubleValue> GetCombinations() {
      var solutions = new List<DoubleValue>();
      double value = ((int)Math.Round(LowerBound.Value / StepSize.Value, 0)) * StepSize.Value;
      if (value < LowerBound.Value) value += StepSize.Value;

      while (value <= UpperBound.Value) {
        solutions.Add(new DoubleValue(value));
        value += StepSize.Value;
      }
      return solutions;
    }
  }
}
